# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module with helper class for elliptic curves."""

# Imports
from typing import Type

from ..common.ikeys import IPrivateKey, IPublicKey
from ..common.ipoint import IPoint


class EllipticCurve:
    """
    Class for a generic elliptic curve.
    This is not meant to be complete but just the minimum required to abstract the bip module from
    the specific ECC library.
    """

    m_name: str
    m_order: int
    m_generator: IPoint
    m_point_cls: Type[IPoint]
    m_pub_key_cls: Type[IPublicKey]
    m_priv_key_cls: Type[IPrivateKey]

    def __init__(self,  # pylint: disable=too-many-arguments
                 name: str,
                 order: int,
                 generator: IPoint,
                 point_cls: Type[IPoint],
                 pub_key_cls: Type[IPublicKey],
                 priv_key_cls: Type[IPrivateKey]):
        """
        Construct class.

        Args:
            name (str)                      : Curve name
            order (int)                     : Curve order
            generator (IPoint object)       : Curve generator point
            point_cls (IPoint class)        : Point class
            pub_key_cls (IPublicKey class)  : Public key class
            priv_key_cls (IPrivateKey class): Private key class
        """
        self.m_name = name
        self.m_order = order
        self.m_generator = generator
        self.m_point_cls = point_cls
        self.m_pub_key_cls = pub_key_cls
        self.m_priv_key_cls = priv_key_cls

    def Name(self) -> str:
        """
        Return the curve name.

        Returns:
            str: Curve name
        """
        return self.m_name

    def Order(self) -> int:
        """
        Return the curve order.

        Returns:
            int: Curve order
        """
        return self.m_order

    def Generator(self) -> IPoint:
        """
        Get the curve generator point.

        Returns:
            IPoint object: IPoint object
        """
        return self.m_generator

    def PointClass(self) -> Type[IPoint]:
        """
        Return the point class.

        Returns:
            IPoint class: Point class
        """
        return self.m_point_cls

    def PublicKeyClass(self) -> Type[IPublicKey]:
        """
        Return the public key class.

        Returns:
            IPublicKey class: Public key class
        """
        return self.m_pub_key_cls

    def PrivateKeyClass(self) -> Type[IPrivateKey]:
        """
        Return the private key class.

        Returns:
            IPrivateKey class: Private key class
        """
        return self.m_priv_key_cls



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for secp256k1 point based on ecdsa library."""

# Imports
from typing import Any

from ....ecdsa import ellipticcurve, keys
from ....ecdsa.ecdsa import curve_secp256k1

from ..common.ipoint import IPoint
from ..curve.elliptic_curve_types import EllipticCurveTypes
from ..ecdsa.ecdsa_keys import EcdsaKeysConst
from ...utils.misc import BytesUtils, DataBytes, IntegerUtils


class Secp256k1PointEcdsa(IPoint):
    """Secp256k1 point class."""

    m_point: ellipticcurve.PointJacobi

    @classmethod
    def FromBytes(cls,
                  point_bytes: bytes) -> IPoint:
        """
        Construct class from point bytes.

        Args:
            point_bytes (bytes): Point bytes

        Returns:
            IPoint: IPoint object
        """
        try:
            return cls(ellipticcurve.PointJacobi.from_bytes(curve_secp256k1,
                                                            point_bytes))
        except keys.MalformedPointError as ex:
            raise ValueError("Invalid point key bytes") from ex
        # ECDSA < 0.17 doesn't have from_bytes method for PointJacobi
        except AttributeError:
            return cls.FromCoordinates(
                BytesUtils.ToInteger(point_bytes[:EcdsaKeysConst.POINT_COORD_BYTE_LEN]),
                BytesUtils.ToInteger(point_bytes[EcdsaKeysConst.POINT_COORD_BYTE_LEN:])
            )

    @classmethod
    def FromCoordinates(cls,
                        x: int,
                        y: int) -> IPoint:
        """
        Construct class from point coordinates.

        Args:
            x (int): X coordinate of the point
            y (int): Y coordinate of the point

        Returns:
            IPoint: IPoint object
        """
        return cls(
            ellipticcurve.PointJacobi.from_affine(
                ellipticcurve.Point(curve_secp256k1, x, y)
            )
        )

    def __init__(self,
                 point_obj: ellipticcurve.PointJacobi) -> None:
        """
        Construct class from point object.

        Args:
            point_obj (ellipticcurve.PointJacobi): Point object
        """
        self.m_point = point_obj

    @staticmethod
    def CurveType() -> EllipticCurveTypes:
        """
        Get the elliptic curve type.

        Returns:
           EllipticCurveTypes: Elliptic curve type
        """
        return EllipticCurveTypes.SECP256K1

    @staticmethod
    def CoordinateLength() -> int:
        """
        Get the coordinate length.

        Returns:
           int: Coordinate key length
        """
        return EcdsaKeysConst.POINT_COORD_BYTE_LEN

    def UnderlyingObject(self) -> Any:
        """
        Get the underlying object.

        Returns:
           Any: Underlying object
        """
        return self.m_point

    def X(self) -> int:
        """
        Get point X coordinate.

        Returns:
           int: Point X coordinate
        """
        return self.m_point.x()

    def Y(self) -> int:
        """
        Get point Y coordinate.

        Returns:
           int: Point Y coordinate
        """
        return self.m_point.y()

    def Raw(self) -> DataBytes:
        """
        Return the point raw bytes.

        Returns:
            DataBytes object: DataBytes object
        """
        return self.RawDecoded()

    def RawEncoded(self) -> DataBytes:
        """
        Return the encoded point raw bytes.

        Returns:
            DataBytes object: DataBytes object
        """
        try:
            return DataBytes(self.m_point.to_bytes("compressed"))
        # ECDSA < 0.17 doesn't have to_bytes method for PointJacobi
        except AttributeError:
            x_bytes = IntegerUtils.ToBytes(self.m_point.x(), EcdsaKeysConst.POINT_COORD_BYTE_LEN)
            if self.m_point.y() & 1:
                enc_bytes = b"\x03" + x_bytes
            else:
                enc_bytes = b"\x02" + x_bytes
            return DataBytes(enc_bytes)

    def RawDecoded(self) -> DataBytes:
        """
        Return the decoded point raw bytes.

        Returns:
            DataBytes object: DataBytes object
        """
        try:
            return DataBytes(self.m_point.to_bytes())
        # ECDSA < 0.17 doesn't have to_bytes method for PointJacobi
        except AttributeError:
            x_bytes = IntegerUtils.ToBytes(self.m_point.x(), EcdsaKeysConst.POINT_COORD_BYTE_LEN)
            y_bytes = IntegerUtils.ToBytes(self.m_point.y(), EcdsaKeysConst.POINT_COORD_BYTE_LEN)

            return DataBytes(x_bytes + y_bytes)

    def __add__(self,
                point: IPoint) -> IPoint:
        """
        Add point to another point.

        Args:
            point (IPoint object): IPoint object

        Returns:
            IPoint object: IPoint object
        """
        return self.__class__(self.m_point + point.UnderlyingObject())

    def __radd__(self,
                 point: IPoint) -> IPoint:
        """
        Add point to another point.

        Args:
            point (IPoint object): IPoint object

        Returns:
            IPoint object: IPoint object
        """
        return self + point

    def __mul__(self,
                scalar: int) -> IPoint:
        """
        Multiply point by a scalar.

        Args:
            scalar (int): scalar

        Returns:
            IPoint object: IPoint object
        """
        return self.__class__(self.m_point * scalar)

    def __rmul__(self,
                 scalar: int) -> IPoint:
        """
        Multiply point by a scalar.

        Args:
            scalar (int): scalar

        Returns:
            IPoint object: IPoint object
        """
        return self * scalar



# -*- coding: utf-8 -*-

import sys
import bisect

"""Author: Carlo Revelli"""
"""Fast bisect bindings"""
"""https://github.com/python/cpython/blob/master/Modules/_bisectmodule.c"""
"""Performs a binary search when inserting keys in sorted order"""


class OrderBookSide(list):
    side = None  # set to True for bids and False for asks

    def __init__(self, deltas=[], depth=None):
        super(OrderBookSide, self).__init__()
        self._depth = depth or sys.maxsize
        self._n = sys.maxsize
        # parallel to self
        self._index = []
        for delta in deltas:
            self.storeArray(list(delta))

    def store_array(self, delta):
        return self.storeArray(delta)

    def storeArray(self, delta):
        price = delta[0]
        size = delta[1]
        index_price = -price if self.side else price
        index = bisect.bisect_left(self._index, index_price)
        if size:
            if index < len(self._index) and self._index[index] == index_price:
                self[index][1] = size
            else:
                self._index.insert(index, index_price)
                self.insert(index, delta)
        elif index < len(self._index) and self._index[index] == index_price:
            del self._index[index]
            del self[index]

    def store(self, price, size):
        self.storeArray([price, size])

    def limit(self):
        difference = len(self) - self._depth
        for _ in range(difference):
            self.remove_index(self.pop())
            self._index.pop()

    def remove_index(self, order):
        pass

    def __len__(self):
        length = super(OrderBookSide, self).__len__()
        return min(length, self._n)

    def __getitem__(self, item):
        if isinstance(item, slice):
            start, stop, step = item.indices(len(self))
            return [self[i] for i in range(start, stop, step)]
        else:
            return super(OrderBookSide, self).__getitem__(item)

    def __eq__(self, other):
        if isinstance(other, list):
            return list(self) == other
        return super(OrderBookSide, self).__eq__(other)

    def __repr__(self):
        return str(list(self))

# -----------------------------------------------------------------------------
# overwrites absolute volumes at price levels
# or deletes price levels based on order counts (3rd value in a bidask delta)
# this class stores vector arrays of values indexed by price


class CountedOrderBookSide(OrderBookSide):
    def __init__(self, deltas=[], depth=None):
        super(CountedOrderBookSide, self).__init__(deltas, depth)

    def storeArray(self, delta):
        price = delta[0]
        size = delta[1]
        count = delta[2]
        index_price = -price if self.side else price
        index = bisect.bisect_left(self._index, index_price)
        if size and count:
            if index < len(self._index) and self._index[index] == index_price:
                self[index][1] = size
                self[index][2] = count
            else:
                self._index.insert(index, index_price)
                self.insert(index, delta)
        elif index < len(self._index) and self._index[index] == index_price:
            del self._index[index]
            del self[index]

    def store(self, price, size, count):
        self.storeArray([price, size, count])

# -----------------------------------------------------------------------------
# indexed by order ids (3rd value in a bidask delta)


class IndexedOrderBookSide(OrderBookSide):
    def __init__(self, deltas=[], depth=None):
        self._hashmap = {}
        super(IndexedOrderBookSide, self).__init__(deltas, depth)

    def storeArray(self, delta):
        price = delta[0]
        if price is not None:
            index_price = -price if self.side else price
        else:
            index_price = None
        size = delta[1]
        order_id = delta[2]
        if size:
            if order_id in self._hashmap:
                old_price = self._hashmap[order_id]
                index_price = index_price or old_price
                # in case the price is not defined
                delta[0] = abs(index_price)
                # matches if price is not defined or if price matches
                if index_price == old_price:
                    # just overwrite the old index
                    index = bisect.bisect_left(self._index, index_price)
                    while self[index][2] != order_id:
                        index += 1
                    self._index[index] = index_price
                    self[index] = delta
                    return
                else:
                    # remove old price level
                    old_index = bisect.bisect_left(self._index, old_price)
                    while self[old_index][2] != order_id:
                        old_index += 1
                    del self._index[old_index]
                    del self[old_index]
            # insert new price level
            self._hashmap[order_id] = index_price
            index = bisect.bisect_left(self._index, index_price)
            while index < len (self._index) and self._index[index] == index_price and self[index][2] < order_id:
                index += 1
            self._index.insert(index, index_price)
            self.insert(index, delta)
        elif order_id in self._hashmap:
            old_price = self._hashmap[order_id]
            index = bisect.bisect_left(self._index, old_price)
            while self[index][2] != order_id:
                index += 1
            del self._index[index]
            del self[index]
            del self._hashmap[order_id]

    def remove_index(self, order):
        order_id = order[2]
        if order_id in self._hashmap:
            del self._hashmap[order_id]

    def store(self, price, size, order_id):
        self.storeArray([price, size, order_id])

# -----------------------------------------------------------------------------
# a more elegant syntax is possible here, but native inheritance is portable

class Asks(OrderBookSide): side = False                                     # noqa
class Bids(OrderBookSide): side = True                                      # noqa
class CountedAsks(CountedOrderBookSide): side = False                       # noqa
class CountedBids(CountedOrderBookSide): side = True                        # noqa
class IndexedAsks(IndexedOrderBookSide): side = False                       # noqa
class IndexedBids(IndexedOrderBookSide): side = True                        # noqa



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for BIP32 paths parsing and handling."""

# Import
from __future__ import annotations

from typing import Iterator, List, Optional, Sequence, Tuple, Union

from .bip32_ex import Bip32PathError
from .bip32_key_data import Bip32KeyIndex


class Bip32PathConst:
    """Class container for BIP32 path constants."""

    # Hardened characters
    HARDENED_CHARS: Tuple[str, str, str] = ("'", "h", "p")
    # Master character
    MASTER_CHAR: str = "m"


class Bip32Path:
    """
    BIP32 path class.
    It represents a BIP-0032 path.
    """

    m_elems: List[Bip32KeyIndex]
    m_is_absolute: bool

    def __init__(self,
                 elems: Optional[Sequence[Union[int, Bip32KeyIndex]]] = None,
                 is_absolute: bool = True) -> None:
        """
        Construct class.

        Args:
            elems (list, optional)      : Path elements (default: empty)
            is_absolute (bool, optional): True if path is an absolute one, false otherwise (default: True)
        """
        try:
            self.m_elems = ([]
                            if elems is None
                            else [Bip32KeyIndex(elem) if isinstance(elem, int) else elem for elem in elems])
        except ValueError as ex:
            raise Bip32PathError("The path contains some invalid key indexes") from ex

        self.m_is_absolute = is_absolute

    def AddElem(self,
                elem: Union[int, Bip32KeyIndex]) -> Bip32Path:
        """
        Return a new path object with the specified element added.

        Args:
            elem (str or Bip32KeyIndex): Path element

        Returns:
            Bip32Path object: Bip32Path object

        Raises:
            Bip32PathError: If the path element is not valid
        """
        if isinstance(elem, int):
            elem = Bip32KeyIndex(elem)
        return Bip32Path(self.m_elems + [elem], self.m_is_absolute)

    def IsAbsolute(self) -> bool:
        """
        Get if absolute path.

        Returns:
            bool: True if absolute path, false otherwise
        """
        return self.m_is_absolute

    def Length(self) -> int:
        """
        Get the number of elements of the path.

        Returns:
            int: Number of elements
        """
        return len(self.m_elems)

    def ToList(self) -> List[int]:
        """
        Get the path as a list of integers.

        Returns:
            list[int]: Path as a list of integers
        """
        return [int(elem) for elem in self.m_elems]

    def ToStr(self) -> str:
        """
        Get the path as a string.

        Returns:
            str: Path as a string
        """
        path_str = "" if not self.m_is_absolute else f"{Bip32PathConst.MASTER_CHAR}/"
        for elem in self.m_elems:
            if not elem.IsHardened():
                path_str += f"{str(elem.ToInt())}/"
            else:
                path_str += f"{str(Bip32KeyIndex.UnhardenIndex(elem.ToInt()))}'/"

        return path_str[:-1]

    def __str__(self) -> str:
        """
        Get the path as a string.

        Returns:
            str: Path as a string
        """
        return self.ToStr()

    def __getitem__(self,
                    idx: int) -> Bip32KeyIndex:
        """
        Get the specified element index.

        Args:
            idx (int): Element index

        Returns:
            Bip32KeyIndex object: Bip32KeyIndex object
        """
        return self.m_elems[idx]

    def __iter__(self) -> Iterator[Bip32KeyIndex]:
        """
        Get the iterator to the current element.

        Returns:
            Iterator object: Iterator to the current element
        """
        yield from self.m_elems


class Bip32PathParser:
    """
    BIP32 path parser class.
    It parses a BIP-0032 path and returns a Bip32Path object.
    """

    @staticmethod
    def Parse(path: str) -> Bip32Path:
        """
        Parse a path and return a Bip32Path object.

        Args:
            path (str): Path

        Returns:
            Bip32Path object: Bip32Path object

        Raises:
            Bip32PathError: If the path is not valid
        """

        # Remove trailing "/" if any
        if path.endswith("/"):
            path = path[:-1]

        # Parse elements
        return Bip32PathParser.__ParseElements(
            list(filter(None, path.split("/")))
        )

    @staticmethod
    def __ParseElements(path_elems: List[str]) -> Bip32Path:
        """
        Parse path elements and return a Bip32Path object.

        Args:
            path_elems (list[str]): Path elements

        Returns:
            Bip32Path object: Bip32Path object

        Raises:
            Bip32PathError: If the path is not valid
        """

        # Remove the initial "m" character if any
        if len(path_elems) > 0 and path_elems[0] == Bip32PathConst.MASTER_CHAR:
            path_elems = path_elems[1:]
            is_absolute = True
        else:
            is_absolute = False

        # Parse elements
        parsed_elems = list(map(Bip32PathParser.__ParseElem, path_elems))
        return Bip32Path(parsed_elems, is_absolute)

    @staticmethod
    def __ParseElem(path_elem: str) -> int:
        """
        Parse path element and get the correspondent index.

        Args:
            path_elem (str): Path element

        Returns:
            int: Index of the element, None if the element is not a valid index

        Raises:
            Bip32PathError: If the path is not valid
        """

        # Strip spaces
        path_elem = path_elem.strip()

        # Get if hardened
        is_hardened = path_elem.endswith(Bip32PathConst.HARDENED_CHARS)

        # If hardened, remove the last character from the string
        if is_hardened:
            path_elem = path_elem[:-1]

        # The remaining string shall be numeric
        if not path_elem.isnumeric():
            raise Bip32PathError(f"Invalid path element ({path_elem})")

        return int(path_elem) if not is_hardened else Bip32KeyIndex.HardenIndex(int(path_elem))



# sympy.external.ntheory
#
# This module provides pure Python implementations of some number theory
# functions that are alternately used from gmpy2 if it is installed.

import sys
import math

# import mpmath.libmp as mlib


_small_trailing = [0] * 256
for j in range(1, 8):
    _small_trailing[1 << j :: 1 << (j + 1)] = [j] * (1 << (7 - j))


# def bit_scan1(x, n=0):
#     if not x:
#         return
#     x = abs(x >> n)
#     low_byte = x & 0xFF
#     if low_byte:
#         return _small_trailing[low_byte] + n

#     t = 8 + n
#     x >>= 8
#     # 2**m is quick for z up through 2**30
#     z = x.bit_length() - 1
#     if x == 1 << z:
#         return z + t

#     if z < 300:
#         # fixed 8-byte reduction
#         while not x & 0xFF:
#             x >>= 8
#             t += 8
#     else:
#         # binary reduction important when there might be a large
#         # number of trailing 0s
#         p = z >> 1
#         while not x & 0xFF:
#             while x & ((1 << p) - 1):
#                 p >>= 1
#             x >>= p
#             t += p
#     return t + _small_trailing[x & 0xFF]


# def bit_scan0(x, n=0):
#     return bit_scan1(x + (1 << n), n)


# def remove(x, f):
#     if f < 2:
#         raise ValueError("factor must be > 1")
#     if x == 0:
#         return 0, 0
#     if f == 2:
#         b = bit_scan1(x)
#         return x >> b, b
#     m = 0
#     y, rem = divmod(x, f)
#     while not rem:
#         x = y
#         m += 1
#         if m > 5:
#             pow_list = [f**2]
#             while pow_list:
#                 _f = pow_list[-1]
#                 y, rem = divmod(x, _f)
#                 if not rem:
#                     m += 1 << len(pow_list)
#                     x = y
#                     pow_list.append(_f**2)
#                 else:
#                     pow_list.pop()
#         y, rem = divmod(x, f)
#     return x, m


# def factorial(x):
#     """Return x!."""
#     return int(mlib.ifac(int(x)))


# def sqrt(x):
#     """Integer square root of x."""
#     return int(mlib.isqrt(int(x)))


# def sqrtrem(x):
#     """Integer square root of x and remainder."""
#     s, r = mlib.sqrtrem(int(x))
#     return (int(s), int(r))


# if sys.version_info[:2] >= (3, 9):
#     # As of Python 3.9 these can take multiple arguments
#     gcd = math.gcd
#     lcm = math.lcm

# else:
#     # Until python 3.8 is no longer supported
#     from functools import reduce


#     def gcd(*args):
#         """gcd of multiple integers."""
#         return reduce(math.gcd, args, 0)


#     def lcm(*args):
#         """lcm of multiple integers."""
#         if 0 in args:
#             return 0
#         return reduce(lambda x, y: x*y//math.gcd(x, y), args, 1)


def _sign(n):
    if n < 0:
        return -1, -n
    return 1, n


def gcdext(a, b):
    if not a or not b:
        g = abs(a) or abs(b)
        if not g:
            return (0, 0, 0)
        return (g, a // g, b // g)

    x_sign, a = _sign(a)
    y_sign, b = _sign(b)
    x, r = 1, 0
    y, s = 0, 1

    while b:
        q, c = divmod(a, b)
        a, b = b, c
        x, r = r, x - q*r
        y, s = s, y - q*s

    return (a, x * x_sign, y * y_sign)


# def is_square(x):
#     """Return True if x is a square number."""
#     if x < 0:
#         return False

#     # Note that the possible values of y**2 % n for a given n are limited.
#     # For example, when n=4, y**2 % n can only take 0 or 1.
#     # In other words, if x % 4 is 2 or 3, then x is not a square number.
#     # Mathematically, it determines if it belongs to the set {y**2 % n},
#     # but implementationally, it can be realized as a logical conjunction
#     # with an n-bit integer.
#     # see https://mersenneforum.org/showpost.php?p=110896
#     # def magic(n):
#     #     s = {y**2 % n for y in range(n)}
#     #     s = set(range(n)) - s
#     #     return sum(1 << bit for bit in s)
#     # >>> print(hex(magic(128)))
#     # 0xfdfdfdedfdfdfdecfdfdfdedfdfcfdec
#     # >>> print(hex(magic(99)))
#     # 0x5f6f9ffb6fb7ddfcb75befdec
#     # >>> print(hex(magic(91)))
#     # 0x6fd1bfcfed5f3679d3ebdec
#     # >>> print(hex(magic(85)))
#     # 0xdef9ae771ffe3b9d67dec
#     if 0xfdfdfdedfdfdfdecfdfdfdedfdfcfdec & (1 << (x & 127)):
#         return False  # e.g. 2, 3
#     m = x % 765765 # 765765 = 99 * 91 * 85
#     if 0x5f6f9ffb6fb7ddfcb75befdec & (1 << (m % 99)):
#         return False  # e.g. 17, 68
#     if 0x6fd1bfcfed5f3679d3ebdec & (1 << (m % 91)):
#         return False  # e.g. 97, 388
#     if 0xdef9ae771ffe3b9d67dec & (1 << (m % 85)):
#         return False  # e.g. 793, 1408
#     return mlib.sqrtrem(int(x))[1] == 0


# def invert(x, m):
#     """Modular inverse of x modulo m.

#     Returns y such that x*y == 1 mod m.

#     Uses ``math.pow`` but reproduces the behaviour of ``gmpy2.invert``
#     which raises ZeroDivisionError if no inverse exists.
#     """
#     try:
#         return pow(x, -1, m)
#     except ValueError:
#         raise ZeroDivisionError("invert() no inverse exists")


# def legendre(x, y):
#     """Legendre symbol (x / y).

#     Following the implementation of gmpy2,
#     the error is raised only when y is an even number.
#     """
#     if y <= 0 or not y % 2:
#         raise ValueError("y should be an odd prime")
#     x %= y
#     if not x:
#         return 0
#     if pow(x, (y - 1) // 2, y) == 1:
#         return 1
#     return -1


# def jacobi(x, y):
#     """Jacobi symbol (x / y)."""
#     if y <= 0 or not y % 2:
#         raise ValueError("y should be an odd positive integer")
#     x %= y
#     if not x:
#         return int(y == 1)
#     if y == 1 or x == 1:
#         return 1
#     if gcd(x, y) != 1:
#         return 0
#     j = 1
#     while x != 0:
#         while x % 2 == 0 and x > 0:
#             x >>= 1
#             if y % 8 in [3, 5]:
#                 j = -j
#         x, y = y, x
#         if x % 4 == y % 4 == 3:
#             j = -j
#         x %= y
#     return j


# def kronecker(x, y):
#     """Kronecker symbol (x / y)."""
#     if gcd(x, y) != 1:
#         return 0
#     if y == 0:
#         return 1
#     sign = -1 if y < 0 and x < 0 else 1
#     y = abs(y)
#     s = bit_scan1(y)
#     y >>= s
#     if s % 2 and x % 8 in [3, 5]:
#         sign = -sign
#     return sign * jacobi(x, y)


# def iroot(y, n):
#     if y < 0:
#         raise ValueError("y must be nonnegative")
#     if n < 1:
#         raise ValueError("n must be positive")
#     if y in (0, 1):
#         return y, True
#     if n == 1:
#         return y, True
#     if n == 2:
#         x, rem = mlib.sqrtrem(y)
#         return int(x), not rem
#     if n >= y.bit_length():
#         return 1, False
#     # Get initial estimate for Newton's method. Care must be taken to
#     # avoid overflow
#     try:
#         guess = int(y**(1./n) + 0.5)
#     except OverflowError:
#         exp = math.log2(y)/n
#         if exp > 53:
#             shift = int(exp - 53)
#             guess = int(2.0**(exp - shift) + 1) << shift
#         else:
#             guess = int(2.0**exp)
#     if guess > 2**50:
#         # Newton iteration
#         xprev, x = -1, guess
#         while 1:
#             t = x**(n - 1)
#             xprev, x = x, ((n - 1)*x + y//t)//n
#             if abs(x - xprev) < 2:
#                 break
#     else:
#         x = guess
#     # Compensate
#     t = x**n
#     while t < y:
#         x += 1
#         t = x**n
#     while t > y:
#         x -= 1
#         t = x**n
#     return x, t == y


# def is_fermat_prp(n, a):
#     if a < 2:
#         raise ValueError("is_fermat_prp() requires 'a' greater than or equal to 2")
#     if n < 1:
#         raise ValueError("is_fermat_prp() requires 'n' be greater than 0")
#     if n == 1:
#         return False
#     if n % 2 == 0:
#         return n == 2
#     a %= n
#     if gcd(n, a) != 1:
#         raise ValueError("is_fermat_prp() requires gcd(n,a) == 1")
#     return pow(a, n - 1, n) == 1


# def is_euler_prp(n, a):
#     if a < 2:
#         raise ValueError("is_euler_prp() requires 'a' greater than or equal to 2")
#     if n < 1:
#         raise ValueError("is_euler_prp() requires 'n' be greater than 0")
#     if n == 1:
#         return False
#     if n % 2 == 0:
#         return n == 2
#     a %= n
#     if gcd(n, a) != 1:
#         raise ValueError("is_euler_prp() requires gcd(n,a) == 1")
#     return pow(a, n >> 1, n) == jacobi(a, n) % n


# def _is_strong_prp(n, a):
#     s = bit_scan1(n - 1)
#     a = pow(a, n >> s, n)
#     if a == 1 or a == n - 1:
#         return True
#     for _ in range(s - 1):
#         a = pow(a, 2, n)
#         if a == n - 1:
#             return True
#         if a == 1:
#             return False
#     return False


# def is_strong_prp(n, a):
#     if a < 2:
#         raise ValueError("is_strong_prp() requires 'a' greater than or equal to 2")
#     if n < 1:
#         raise ValueError("is_strong_prp() requires 'n' be greater than 0")
#     if n == 1:
#         return False
#     if n % 2 == 0:
#         return n == 2
#     a %= n
#     if gcd(n, a) != 1:
#         raise ValueError("is_strong_prp() requires gcd(n,a) == 1")
#     return _is_strong_prp(n, a)


# def _lucas_sequence(n, P, Q, k):
#     r"""Return the modular Lucas sequence (U_k, V_k, Q_k).

#     Explanation
#     ===========

#     Given a Lucas sequence defined by P, Q, returns the kth values for
#     U and V, along with Q^k, all modulo n. This is intended for use with
#     possibly very large values of n and k, where the combinatorial functions
#     would be completely unusable.

#     .. math ::
#         U_k = \begin{cases}
#              0 & \text{if } k = 0\\
#              1 & \text{if } k = 1\\
#              PU_{k-1} - QU_{k-2} & \text{if } k > 1
#         \end{cases}\\
#         V_k = \begin{cases}
#              2 & \text{if } k = 0\\
#              P & \text{if } k = 1\\
#              PV_{k-1} - QV_{k-2} & \text{if } k > 1
#         \end{cases}

#     The modular Lucas sequences are used in numerous places in number theory,
#     especially in the Lucas compositeness tests and the various n + 1 proofs.

#     Parameters
#     ==========

#     n : int
#         n is an odd number greater than or equal to 3
#     P : int
#     Q : int
#         D determined by D = P**2 - 4*Q is non-zero
#     k : int
#         k is a nonnegative integer

#     Returns
#     =======

#     U, V, Qk : (int, int, int)
#         `(U_k \bmod{n}, V_k \bmod{n}, Q^k \bmod{n})`

#     Examples
#     ========

#     >>> from sympy.external.ntheory import _lucas_sequence
#     >>> N = 10**2000 + 4561
#     >>> sol = U, V, Qk = _lucas_sequence(N, 3, 1, N//2); sol
#     (0, 2, 1)

#     References
#     ==========

#     .. [1] https://en.wikipedia.org/wiki/Lucas_sequence

#     """
#     if k == 0:
#         return (0, 2, 1)
#     D = P**2 - 4*Q
#     U = 1
#     V = P
#     Qk = Q % n
#     if Q == 1:
#         # Optimization for extra strong tests.
#         for b in bin(k)[3:]:
#             U = (U*V) % n
#             V = (V*V - 2) % n
#             if b == "1":
#                 U, V = U*P + V, V*P + U*D
#                 if U & 1:
#                     U += n
#                 if V & 1:
#                     V += n
#                 U, V = U >> 1, V >> 1
#     elif P == 1 and Q == -1:
#         # Small optimization for 50% of Selfridge parameters.
#         for b in bin(k)[3:]:
#             U = (U*V) % n
#             if Qk == 1:
#                 V = (V*V - 2) % n
#             else:
#                 V = (V*V + 2) % n
#                 Qk = 1
#             if b == "1":
#                 # new_U = (U + V) // 2
#                 # new_V = (5*U + V) // 2 = 2*U + new_U
#                 U, V  = U + V, U << 1
#                 if U & 1:
#                     U += n
#                 U >>= 1
#                 V += U
#                 Qk = -1
#         Qk %= n
#     elif P == 1:
#         for b in bin(k)[3:]:
#             U = (U*V) % n
#             V = (V*V - 2*Qk) % n
#             Qk *= Qk
#             if b == "1":
#                 # new_U = (U + V) // 2
#                 # new_V = new_U - 2*Q*U
#                 U, V  = U + V, (Q*U) << 1
#                 if U & 1:
#                     U += n
#                 U >>= 1
#                 V = U - V
#                 Qk *= Q
#             Qk %= n
#     else:
#         # The general case with any P and Q.
#         for b in bin(k)[3:]:
#             U = (U*V) % n
#             V = (V*V - 2*Qk) % n
#             Qk *= Qk
#             if b == "1":
#                 U, V = U*P + V, V*P + U*D
#                 if U & 1:
#                     U += n
#                 if V & 1:
#                     V += n
#                 U, V = U >> 1, V >> 1
#                 Qk *= Q
#             Qk %= n
#     return (U % n, V % n, Qk)


# def is_fibonacci_prp(n, p, q):
#     d = p**2 - 4*q
#     if d == 0 or p <= 0 or q not in [1, -1]:
#         raise ValueError("invalid values for p,q in is_fibonacci_prp()")
#     if n < 1:
#         raise ValueError("is_fibonacci_prp() requires 'n' be greater than 0")
#     if n == 1:
#         return False
#     if n % 2 == 0:
#         return n == 2
#     return _lucas_sequence(n, p, q, n)[1] == p % n


# def is_lucas_prp(n, p, q):
#     d = p**2 - 4*q
#     if d == 0:
#         raise ValueError("invalid values for p,q in is_lucas_prp()")
#     if n < 1:
#         raise ValueError("is_lucas_prp() requires 'n' be greater than 0")
#     if n == 1:
#         return False
#     if n % 2 == 0:
#         return n == 2
#     if gcd(n, q*d) not in [1, n]:
#         raise ValueError("is_lucas_prp() requires gcd(n,2*q*D) == 1")
#     return _lucas_sequence(n, p, q, n - jacobi(d, n))[0] == 0


# def _is_selfridge_prp(n):
#     """Lucas compositeness test with the Selfridge parameters for n.

#     Explanation
#     ===========

#     The Lucas compositeness test checks whether n is a prime number.
#     The test can be run with arbitrary parameters ``P`` and ``Q``, which also change the performance of the test.
#     So, which parameters are most effective for running the Lucas compositeness test?
#     As an algorithm for determining ``P`` and ``Q``, Selfridge proposed method A [1]_ page 1401
#     (Since two methods were proposed, referred to simply as A and B in the paper,
#     we will refer to one of them as "method A").

#     method A fixes ``P = 1``. Then, ``D`` defined by ``D = P**2 - 4Q`` is varied from 5, -7, 9, -11, 13, and so on,
#     with the first ``D`` being ``jacobi(D, n) == -1``. Once ``D`` is determined,
#     ``Q`` is determined to be ``(P**2 - D)//4``.

#     References
#     ==========

#     .. [1] Robert Baillie, Samuel S. Wagstaff, Lucas Pseudoprimes,
#            Math. Comp. Vol 35, Number 152 (1980), pp. 1391-1417,
#            https://doi.org/10.1090%2FS0025-5718-1980-0583518-6
#            http://mpqs.free.fr/LucasPseudoprimes.pdf

#     """
#     for D in range(5, 1_000_000, 2):
#         if D & 2: # if D % 4 == 3
#             D = -D
#         j = jacobi(D, n)
#         if j == -1:
#             return _lucas_sequence(n, 1, (1-D) // 4, n + 1)[0] == 0
#         if j == 0 and D % n:
#             return False
#         # When j == -1 is hard to find, suspect a square number
#         if D == 13 and is_square(n):
#             return False
#     raise ValueError("appropriate value for D cannot be found in is_selfridge_prp()")


# def is_selfridge_prp(n):
#     if n < 1:
#         raise ValueError("is_selfridge_prp() requires 'n' be greater than 0")
#     if n == 1:
#         return False
#     if n % 2 == 0:
#         return n == 2
#     return _is_selfridge_prp(n)


# def is_strong_lucas_prp(n, p, q):
#     D = p**2 - 4*q
#     if D == 0:
#         raise ValueError("invalid values for p,q in is_strong_lucas_prp()")
#     if n < 1:
#         raise ValueError("is_selfridge_prp() requires 'n' be greater than 0")
#     if n == 1:
#         return False
#     if n % 2 == 0:
#         return n == 2
#     if gcd(n, q*D) not in [1, n]:
#         raise ValueError("is_strong_lucas_prp() requires gcd(n,2*q*D) == 1")
#     j = jacobi(D, n)
#     s = bit_scan1(n - j)
#     U, V, Qk = _lucas_sequence(n, p, q, (n - j) >> s)
#     if U == 0 or V == 0:
#         return True
#     for _ in range(s - 1):
#         V = (V*V - 2*Qk) % n
#         if V == 0:
#             return True
#         Qk = pow(Qk, 2, n)
#     return False


# def _is_strong_selfridge_prp(n):
#     for D in range(5, 1_000_000, 2):
#         if D & 2: # if D % 4 == 3
#             D = -D
#         j = jacobi(D, n)
#         if j == -1:
#             s = bit_scan1(n + 1)
#             U, V, Qk = _lucas_sequence(n, 1, (1-D) // 4, (n + 1) >> s)
#             if U == 0 or V == 0:
#                 return True
#             for _ in range(s - 1):
#                 V = (V*V - 2*Qk) % n
#                 if V == 0:
#                     return True
#                 Qk = pow(Qk, 2, n)
#             return False
#         if j == 0 and D % n:
#             return False
#         # When j == -1 is hard to find, suspect a square number
#         if D == 13 and is_square(n):
#             return False
#     raise ValueError("appropriate value for D cannot be found in is_strong_selfridge_prp()")


# def is_strong_selfridge_prp(n):
#     if n < 1:
#         raise ValueError("is_strong_selfridge_prp() requires 'n' be greater than 0")
#     if n == 1:
#         return False
#     if n % 2 == 0:
#         return n == 2
#     return _is_strong_selfridge_prp(n)


# def is_bpsw_prp(n):
#     if n < 1:
#         raise ValueError("is_bpsw_prp() requires 'n' be greater than 0")
#     if n == 1:
#         return False
#     if n % 2 == 0:
#         return n == 2
#     return _is_strong_prp(n, 2) and _is_selfridge_prp(n)


# def is_strong_bpsw_prp(n):
#     if n < 1:
#         raise ValueError("is_strong_bpsw_prp() requires 'n' be greater than 0")
#     if n == 1:
#         return False
#     if n % 2 == 0:
#         return n == 2
#     return _is_strong_prp(n, 2) and _is_strong_selfridge_prp(n)



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module with secp256k1 constants."""

# Imports
from typing import Type

from ..common.ikeys import IPrivateKey, IPublicKey
from ..common.ipoint import IPoint


# Variables
Secp256k1Point: Type[IPoint]
Secp256k1PublicKey: Type[IPublicKey]
Secp256k1PrivateKey: Type[IPrivateKey]
_CURVE_ORDER: int
_GENERATOR: IPoint

from ....ecdsa.ecdsa import generator_secp256k1

from ..secp256k1.secp256k1_keys_ecdsa import (
    Secp256k1PointEcdsa, Secp256k1PrivateKeyEcdsa, Secp256k1PublicKeyEcdsa
)

Secp256k1Point = Secp256k1PointEcdsa
Secp256k1PublicKey = Secp256k1PublicKeyEcdsa
Secp256k1PrivateKey = Secp256k1PrivateKeyEcdsa

_CURVE_ORDER = generator_secp256k1.order()
_GENERATOR = Secp256k1Point(generator_secp256k1)


class Secp256k1Const:
    """Class container for Secp256k1 constants."""

    # Curve name
    NAME: str = "Secp256k1"
    # Curve order
    CURVE_ORDER: int = _CURVE_ORDER
    # Curve generator point
    GENERATOR: IPoint = _GENERATOR



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module with helper class for generic coins configuration handling."""

# Imports
from typing import Any, Dict

from ..utils.conf import CoinNames as ConfCoinNames


class CoinConf:
    """Coin configuration class."""

    m_coin_name: ConfCoinNames
    m_params: Dict[str, Any]

    def __init__(self,
                 coin_name: ConfCoinNames,
                 params: Dict[str, Any]) -> None:
        """
        Construct class.

        Args:
            coin_name (CoinNames object): Coin names
            params (dict)               : SS58 format
        """
        self.m_coin_name = coin_name
        self.m_params = params

    def CoinNames(self) -> ConfCoinNames:
        """
        Get coin names.

        Returns:
            CoinNames object: CoinNames object
        """
        return self.m_coin_name

    def ParamByKey(self,
                   key: str) -> Any:
        """
        Get the parameter by key.

        Args:
            key (str): Parameter key

        Returns:
            Any: Parameter value
        """
        return self.m_params[key]



# Copyright (c) 2022 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for ChaCha20-Poly1305 algorithm."""

# Imports
from typing import Tuple, Union

from Crypto.Cipher import ChaCha20_Poly1305

from ..misc import AlgoUtils


class ChaCha20Poly1305:
    """
    ChaCha20-Poly1305 class.
    It decrypts/encrypts data using ChaCha20-Poly1305 algorithm.
    """

    @staticmethod
    def Decrypt(key: Union[bytes, str],
                nonce: Union[bytes, str],
                assoc_data: Union[bytes, str],
                cipher_text: Union[bytes, str],
                tag: Union[bytes, str]) -> bytes:
        """
        Decrypt data.

        Args:
            key (str or bytes)       : Key
            nonce (str or bytes)     : Nonce
            assoc_data (str or bytes): Associated data
            cipher_text (bytes)      : Cipher text
            tag (bytes)              : Tag

        Returns:
            bytes: Decrypted data
        """
        cipher = ChaCha20_Poly1305.new(key=AlgoUtils.Encode(key),
                                       nonce=AlgoUtils.Encode(nonce))
        cipher.update(AlgoUtils.Encode(assoc_data))
        return cipher.decrypt_and_verify(AlgoUtils.Encode(cipher_text), AlgoUtils.Encode(tag))

    @staticmethod
    def Encrypt(key: Union[bytes, str],
                nonce: Union[bytes, str],
                assoc_data: Union[bytes, str],
                plain_text: Union[bytes, str]) -> Tuple[bytes, bytes]:
        """
        Encrypt data.

        Args:
            key (str or bytes)       : Key
            nonce (str or bytes)     : Nonce
            assoc_data (str or bytes): Associated data
            plain_text (str or bytes): Plain text

        Returns:
            tuple[bytes, bytes]: Cipher text bytes (index 0) and tag bytes (index 1)
        """
        cipher = ChaCha20_Poly1305.new(key=AlgoUtils.Encode(key),
                                       nonce=AlgoUtils.Encode(nonce))
        cipher.update(AlgoUtils.Encode(assoc_data))
        return cipher.encrypt_and_digest(AlgoUtils.Encode(plain_text))

    @staticmethod
    def KeySize() -> int:
        """
        Get the key size.

        Returns:
            int: Key size
        """
        return ChaCha20_Poly1305.key_size

    @staticmethod
    def TagSize() -> int:
        """
        Get the tag size.

        Returns:
            int: Tag size
        """
        return 16



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module for BIP49 coins configuration."""

# Imports
from bip_utils.addr import BchP2SHAddrEncoder, P2SHAddrEncoder
from bip_utils.bip.bip32 import Bip32KeyNetVersions, Bip32Slip10Secp256k1
from  import DER_PATH_NON_HARDENED_FULL, BipBitcoinCashConf, BipCoinConf, BipLitecoinConf
from bip_utils.coin_conf import CoinsConf
from bip_utils.slip.slip44 import Slip44


# Bitcoin key net version for main net (ypub / yprv)
_BIP49_BTC_KEY_NET_VER_MAIN: Bip32KeyNetVersions = Bip32KeyNetVersions(b"\x04\x9d\x7c\xb2",
                                                                       b"\x04\x9d\x78\x78")
# Bitcoin key net version for test net (upub / uprv)
_BIP49_BTC_KEY_NET_VER_TEST: Bip32KeyNetVersions = Bip32KeyNetVersions(b"\x04\x4a\x52\x62",
                                                                       b"\x04\x4a\x4e\x28")


class Bip49Conf:
    """Class container for BIP49 configuration."""

    # Configuration for Bitcoin main net
    BitcoinMainNet: BipCoinConf = BipCoinConf(
        coin_names=CoinsConf.BitcoinMainNet.CoinNames(),
        coin_idx=Slip44.BITCOIN,
        is_testnet=False,
        def_path=DER_PATH_NON_HARDENED_FULL,
        key_net_ver=_BIP49_BTC_KEY_NET_VER_MAIN,
        wif_net_ver=CoinsConf.BitcoinMainNet.ParamByKey("wif_net_ver"),
        bip32_cls=Bip32Slip10Secp256k1,
        addr_cls=P2SHAddrEncoder,
        addr_params={
            "net_ver": CoinsConf.BitcoinMainNet.ParamByKey("p2sh_net_ver"),
        },
    )
    # Configuration for Bitcoin regtest
    BitcoinRegTest: BipCoinConf = BipCoinConf(
        coin_names=CoinsConf.BitcoinRegTest.CoinNames(),
        coin_idx=Slip44.TESTNET,
        is_testnet=True,
        def_path=DER_PATH_NON_HARDENED_FULL,
        key_net_ver=_BIP49_BTC_KEY_NET_VER_TEST,
        wif_net_ver=CoinsConf.BitcoinRegTest.ParamByKey("wif_net_ver"),
        bip32_cls=Bip32Slip10Secp256k1,
        addr_cls=P2SHAddrEncoder,
        addr_params={
            "net_ver": CoinsConf.BitcoinRegTest.ParamByKey("p2sh_net_ver"),
        },
    )
    # Configuration for Bitcoin test net
    BitcoinTestNet: BipCoinConf = BipCoinConf(
        coin_names=CoinsConf.BitcoinTestNet.CoinNames(),
        coin_idx=Slip44.TESTNET,
        is_testnet=True,
        def_path=DER_PATH_NON_HARDENED_FULL,
        key_net_ver=_BIP49_BTC_KEY_NET_VER_TEST,
        wif_net_ver=CoinsConf.BitcoinTestNet.ParamByKey("wif_net_ver"),
        bip32_cls=Bip32Slip10Secp256k1,
        addr_cls=P2SHAddrEncoder,
        addr_params={
            "net_ver": CoinsConf.BitcoinTestNet.ParamByKey("p2sh_net_ver"),
        },
    )

    # Configuration for Bitcoin Cash main net
    BitcoinCashMainNet: BipBitcoinCashConf = BipBitcoinCashConf(
        coin_names=CoinsConf.BitcoinCashMainNet.CoinNames(),
        coin_idx=Slip44.BITCOIN_CASH,
        is_testnet=False,
        def_path=DER_PATH_NON_HARDENED_FULL,
        key_net_ver=_BIP49_BTC_KEY_NET_VER_MAIN,
        wif_net_ver=CoinsConf.BitcoinCashMainNet.ParamByKey("wif_net_ver"),
        bip32_cls=Bip32Slip10Secp256k1,
        addr_cls=BchP2SHAddrEncoder,
        addr_params={
            "std": {
                "net_ver": CoinsConf.BitcoinCashMainNet.ParamByKey("p2sh_std_net_ver"),
                "hrp": CoinsConf.BitcoinCashMainNet.ParamByKey("p2sh_std_hrp"),
            },
            "legacy": {
                "net_ver": CoinsConf.BitcoinCashMainNet.ParamByKey("p2sh_legacy_net_ver"),
            }
        },
        addr_cls_legacy=P2SHAddrEncoder,
    )
    # Configuration for Bitcoin Cash test net
    BitcoinCashTestNet: BipBitcoinCashConf = BipBitcoinCashConf(
        coin_names=CoinsConf.BitcoinCashTestNet.CoinNames(),
        coin_idx=Slip44.TESTNET,
        is_testnet=True,
        def_path=DER_PATH_NON_HARDENED_FULL,
        key_net_ver=_BIP49_BTC_KEY_NET_VER_TEST,
        wif_net_ver=CoinsConf.BitcoinCashTestNet.ParamByKey("wif_net_ver"),
        bip32_cls=Bip32Slip10Secp256k1,
        addr_cls=BchP2SHAddrEncoder,
        addr_params={
            "std": {
                "net_ver": CoinsConf.BitcoinCashTestNet.ParamByKey("p2sh_std_net_ver"),
                "hrp": CoinsConf.BitcoinCashTestNet.ParamByKey("p2sh_std_hrp"),
            },
            "legacy": {
                "net_ver": CoinsConf.BitcoinCashTestNet.ParamByKey("p2sh_legacy_net_ver"),
            }
        },
        addr_cls_legacy=P2SHAddrEncoder,
    )

    # Configuration for Bitcoin Cash Simple Ledger Protocol main net
    BitcoinCashSlpMainNet: BipBitcoinCashConf = BipBitcoinCashConf(
        coin_names=CoinsConf.BitcoinCashSlpMainNet.CoinNames(),
        coin_idx=Slip44.BITCOIN_CASH,
        is_testnet=False,
        def_path=DER_PATH_NON_HARDENED_FULL,
        key_net_ver=_BIP49_BTC_KEY_NET_VER_MAIN,
        wif_net_ver=CoinsConf.BitcoinCashSlpMainNet.ParamByKey("wif_net_ver"),
        bip32_cls=Bip32Slip10Secp256k1,
        addr_cls=BchP2SHAddrEncoder,
        addr_params={
            "std": {
                "net_ver": CoinsConf.BitcoinCashSlpMainNet.ParamByKey("p2sh_std_net_ver"),
                "hrp": CoinsConf.BitcoinCashSlpMainNet.ParamByKey("p2sh_std_hrp"),
            },
            "legacy": {
                "net_ver": CoinsConf.BitcoinCashSlpMainNet.ParamByKey("p2sh_legacy_net_ver"),
            }
        },
        addr_cls_legacy=P2SHAddrEncoder,
    )
    # Configuration for Bitcoin Cash Simple Ledger Protocol test net
    BitcoinCashSlpTestNet: BipBitcoinCashConf = BipBitcoinCashConf(
        coin_names=CoinsConf.BitcoinCashSlpTestNet.CoinNames(),
        coin_idx=Slip44.TESTNET,
        is_testnet=True,
        def_path=DER_PATH_NON_HARDENED_FULL,
        key_net_ver=_BIP49_BTC_KEY_NET_VER_TEST,
        wif_net_ver=CoinsConf.BitcoinCashSlpTestNet.ParamByKey("wif_net_ver"),
        bip32_cls=Bip32Slip10Secp256k1,
        addr_cls=BchP2SHAddrEncoder,
        addr_params={
            "std": {
                "net_ver": CoinsConf.BitcoinCashSlpTestNet.ParamByKey("p2sh_std_net_ver"),
                "hrp": CoinsConf.BitcoinCashSlpTestNet.ParamByKey("p2sh_std_hrp"),
            },
            "legacy": {
                "net_ver": CoinsConf.BitcoinCashSlpTestNet.ParamByKey("p2sh_legacy_net_ver"),
            }
        },
        addr_cls_legacy=P2SHAddrEncoder,
    )

    # Configuration for BitcoinSV main net
    BitcoinSvMainNet: BipCoinConf = BipCoinConf(
        coin_names=CoinsConf.BitcoinSvMainNet.CoinNames(),
        coin_idx=Slip44.BITCOIN_SV,
        is_testnet=False,
        def_path=DER_PATH_NON_HARDENED_FULL,
        key_net_ver=_BIP49_BTC_KEY_NET_VER_MAIN,
        wif_net_ver=CoinsConf.BitcoinSvMainNet.ParamByKey("wif_net_ver"),
        bip32_cls=Bip32Slip10Secp256k1,
        addr_cls=P2SHAddrEncoder,
        addr_params={
            "net_ver": CoinsConf.BitcoinSvMainNet.ParamByKey("p2sh_net_ver"),
        },
    )
    # Configuration for BitcoinSV test net
    BitcoinSvTestNet: BipCoinConf = BipCoinConf(
        coin_names=CoinsConf.BitcoinSvTestNet.CoinNames(),
        coin_idx=Slip44.TESTNET,
        is_testnet=True,
        def_path=DER_PATH_NON_HARDENED_FULL,
        key_net_ver=_BIP49_BTC_KEY_NET_VER_TEST,
        wif_net_ver=CoinsConf.BitcoinSvTestNet.ParamByKey("wif_net_ver"),
        bip32_cls=Bip32Slip10Secp256k1,
        addr_cls=P2SHAddrEncoder,
        addr_params={
            "net_ver": CoinsConf.BitcoinSvTestNet.ParamByKey("p2sh_net_ver"),
        },
    )

    # Configuration for Dash main net
    DashMainNet: BipCoinConf = BipCoinConf(
        coin_names=CoinsConf.DashMainNet.CoinNames(),
        coin_idx=Slip44.DASH,
        is_testnet=False,
        def_path=DER_PATH_NON_HARDENED_FULL,
        key_net_ver=_BIP49_BTC_KEY_NET_VER_MAIN,
        wif_net_ver=CoinsConf.DashMainNet.ParamByKey("wif_net_ver"),
        bip32_cls=Bip32Slip10Secp256k1,
        addr_cls=P2SHAddrEncoder,
        addr_params={
            "net_ver": CoinsConf.DashMainNet.ParamByKey("p2sh_net_ver"),
        },
    )
    # Configuration for Dash test net
    DashTestNet: BipCoinConf = BipCoinConf(
        coin_names=CoinsConf.DashTestNet.CoinNames(),
        coin_idx=Slip44.TESTNET,
        is_testnet=True,
        def_path=DER_PATH_NON_HARDENED_FULL,
        key_net_ver=_BIP49_BTC_KEY_NET_VER_TEST,
        wif_net_ver=CoinsConf.DashTestNet.ParamByKey("wif_net_ver"),
        bip32_cls=Bip32Slip10Secp256k1,
        addr_cls=P2SHAddrEncoder,
        addr_params={
            "net_ver": CoinsConf.DashTestNet.ParamByKey("p2sh_net_ver"),
        },
    )

    # Configuration for Dogecoin main net
    DogecoinMainNet: BipCoinConf = BipCoinConf(
        coin_names=CoinsConf.DogecoinMainNet.CoinNames(),
        coin_idx=Slip44.DOGECOIN,
        is_testnet=False,
        def_path=DER_PATH_NON_HARDENED_FULL,
        key_net_ver=Bip32KeyNetVersions(b"\x02\xfa\xca\xfd",
                                        b"\x02\xfa\xc3\x98"),   # dgub / dgpv
        wif_net_ver=CoinsConf.DogecoinMainNet.ParamByKey("wif_net_ver"),
        bip32_cls=Bip32Slip10Secp256k1,
        addr_cls=P2SHAddrEncoder,
        addr_params={
            "net_ver": CoinsConf.DogecoinMainNet.ParamByKey("p2sh_net_ver"),
        },
    )
    # Configuration for Dogecoin test net
    DogecoinTestNet: BipCoinConf = BipCoinConf(
        coin_names=CoinsConf.DogecoinTestNet.CoinNames(),
        coin_idx=Slip44.TESTNET,
        is_testnet=True,
        def_path=DER_PATH_NON_HARDENED_FULL,
        key_net_ver=Bip32KeyNetVersions(b"\x04\x32\xa9\xa8",
                                        b"\x04\x32\xa2\x43"),   # tgub / tgpv
        wif_net_ver=CoinsConf.DogecoinTestNet.ParamByKey("wif_net_ver"),
        bip32_cls=Bip32Slip10Secp256k1,
        addr_cls=P2SHAddrEncoder,
        addr_params={
            "net_ver": CoinsConf.DogecoinTestNet.ParamByKey("p2sh_net_ver"),
        },
    )

    # Configuration for eCash main net
    EcashMainNet: BipBitcoinCashConf = BipBitcoinCashConf(
        coin_names=CoinsConf.EcashMainNet.CoinNames(),
        coin_idx=Slip44.BITCOIN_CASH,
        is_testnet=False,
        def_path=DER_PATH_NON_HARDENED_FULL,
        key_net_ver=_BIP49_BTC_KEY_NET_VER_MAIN,
        wif_net_ver=CoinsConf.EcashMainNet.ParamByKey("wif_net_ver"),
        bip32_cls=Bip32Slip10Secp256k1,
        addr_cls=BchP2SHAddrEncoder,
        addr_params={
            "std": {
                "net_ver": CoinsConf.EcashMainNet.ParamByKey("p2sh_std_net_ver"),
                "hrp": CoinsConf.EcashMainNet.ParamByKey("p2sh_std_hrp"),
            },
            "legacy": {
                "net_ver": CoinsConf.EcashMainNet.ParamByKey("p2sh_legacy_net_ver"),
            }
        },
        addr_cls_legacy=P2SHAddrEncoder,
    )
    # Configuration for eCash test net
    EcashTestNet: BipBitcoinCashConf = BipBitcoinCashConf(
        coin_names=CoinsConf.EcashTestNet.CoinNames(),
        coin_idx=Slip44.TESTNET,
        is_testnet=True,
        def_path=DER_PATH_NON_HARDENED_FULL,
        key_net_ver=_BIP49_BTC_KEY_NET_VER_TEST,
        wif_net_ver=CoinsConf.EcashTestNet.ParamByKey("wif_net_ver"),
        bip32_cls=Bip32Slip10Secp256k1,
        addr_cls=BchP2SHAddrEncoder,
        addr_params={
            "std": {
                "net_ver": CoinsConf.EcashTestNet.ParamByKey("p2sh_std_net_ver"),
                "hrp": CoinsConf.EcashTestNet.ParamByKey("p2sh_std_hrp"),
            },
            "legacy": {
                "net_ver": CoinsConf.EcashTestNet.ParamByKey("p2sh_legacy_net_ver"),
            }
        },
        addr_cls_legacy=P2SHAddrEncoder,
    )

    # Configuration for Litecoin main net
    LitecoinMainNet: BipLitecoinConf = BipLitecoinConf(
        coin_names=CoinsConf.LitecoinMainNet.CoinNames(),
        coin_idx=Slip44.LITECOIN,
        is_testnet=False,
        def_path=DER_PATH_NON_HARDENED_FULL,
        key_net_ver=_BIP49_BTC_KEY_NET_VER_MAIN,
        alt_key_net_ver=Bip32KeyNetVersions(b"\x01\xb2\x6e\xf6",
                                            b"\x01\xb2\x67\x92"),   # Mtpv / Mtub
        wif_net_ver=CoinsConf.LitecoinMainNet.ParamByKey("wif_net_ver"),
        bip32_cls=Bip32Slip10Secp256k1,
        addr_cls=P2SHAddrEncoder,
        addr_params={
            "std_net_ver": CoinsConf.LitecoinMainNet.ParamByKey("p2sh_std_net_ver"),
            "depr_net_ver": CoinsConf.LitecoinMainNet.ParamByKey("p2sh_depr_net_ver"),
        },
    )
    # Configuration for Litecoin test net
    LitecoinTestNet: BipLitecoinConf = BipLitecoinConf(
        coin_names=CoinsConf.LitecoinTestNet.CoinNames(),
        coin_idx=Slip44.TESTNET,
        is_testnet=True,
        def_path=DER_PATH_NON_HARDENED_FULL,
        key_net_ver=Bip32KeyNetVersions(b"\x04\x36\xf6\xe1",
                                        b"\x04\x36\xef\x7d"),       # ttub / ttpv
        alt_key_net_ver=Bip32KeyNetVersions(b"\x04\x36\xf6\xe1",
                                            b"\x04\x36\xef\x7d"),   # ttub / ttpv
        wif_net_ver=CoinsConf.LitecoinTestNet.ParamByKey("wif_net_ver"),
        bip32_cls=Bip32Slip10Secp256k1,
        addr_cls=P2SHAddrEncoder,
        addr_params={
            "std_net_ver": CoinsConf.LitecoinTestNet.ParamByKey("p2sh_std_net_ver"),
            "depr_net_ver": CoinsConf.LitecoinTestNet.ParamByKey("p2sh_depr_net_ver"),
        },
    )

    # Configuration for Zcash main net
    ZcashMainNet: BipCoinConf = BipCoinConf(
        coin_names=CoinsConf.ZcashMainNet.CoinNames(),
        coin_idx=Slip44.ZCASH,
        is_testnet=False,
        def_path=DER_PATH_NON_HARDENED_FULL,
        key_net_ver=_BIP49_BTC_KEY_NET_VER_MAIN,
        wif_net_ver=CoinsConf.ZcashMainNet.ParamByKey("wif_net_ver"),
        bip32_cls=Bip32Slip10Secp256k1,
        addr_cls=P2SHAddrEncoder,
        addr_params={
            "net_ver": CoinsConf.ZcashMainNet.ParamByKey("p2sh_net_ver"),
        },
    )
    # Configuration for Zcash test net
    ZcashTestNet: BipCoinConf = BipCoinConf(
        coin_names=CoinsConf.ZcashTestNet.CoinNames(),
        coin_idx=Slip44.TESTNET,
        is_testnet=True,
        def_path=DER_PATH_NON_HARDENED_FULL,
        key_net_ver=_BIP49_BTC_KEY_NET_VER_TEST,
        wif_net_ver=CoinsConf.ZcashTestNet.ParamByKey("wif_net_ver"),
        bip32_cls=Bip32Slip10Secp256k1,
        addr_cls=P2SHAddrEncoder,
        addr_params={
            "net_ver": CoinsConf.ZcashTestNet.ParamByKey("p2sh_net_ver"),
        },
    )



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Module for segwit bech32/bech32m decoding/encoding.

References:
    https://github.com/bitcoin/bips/blob/master/bip-0173.mediawiki
    https://github.com/bitcoin/bips/blob/master/bip-0350.mediawiki
"""

# Imports
from typing import List, Tuple

from .bech32 import Bech32Const, Bech32Encodings, Bech32Utils
from .bech32_base import Bech32BaseUtils, Bech32DecoderBase, Bech32EncoderBase
from ..utils.misc import BytesUtils


class SegwitBech32Const:
    """Class container for Segwit Bech32 constants."""

    # Separator
    SEPARATOR: str = Bech32Const.SEPARATOR
    # Checksum length
    CHECKSUM_STR_LEN: int = Bech32Const.CHECKSUM_STR_LEN
    # Minimum witness program length in bytes
    WITNESS_PROG_MIN_BYTE_LEN: int = 2
    # Maximum witness program length in bytes
    WITNESS_PROG_MAX_BYTE_LEN: int = 40
    # Witness version for Bech32 encoding
    WITNESS_VER_BECH32: int = 0
    # Witness version maximum value
    WITNESS_VER_MAX_VAL: int = 16
    # Accepted data lengths when witness version is zero
    WITNESS_VER_ZERO_DATA_BYTE_LEN: Tuple[int, int] = (20, 32)


class SegwitBech32Encoder(Bech32EncoderBase):
    """
    Segwit Bech32 encoder class.
    It provides methods for encoding to Segwit Bech32 format.
    """

    @classmethod
    def Encode(cls,
               hrp: str,
               wit_ver: int,
               wit_prog: bytes) -> str:
        """
        Encode to Segwit Bech32.

        Args:
            hrp (str)       : HRP
            wit_ver (int)   : Witness version
            wit_prog (bytes): Witness program

        Returns:
            str: Encoded address

        Raises:
            ValueError: If the data is not valid
        """
        return cls._EncodeBech32(hrp,
                                 [wit_ver] + Bech32BaseUtils.ConvertToBase32(wit_prog),
                                 SegwitBech32Const.SEPARATOR)

    @staticmethod
    def _ComputeChecksum(hrp: str,
                         data: List[int]) -> List[int]:
        """
        Compute the checksum from the specified HRP and data.

        Args:
            hrp (str)       : HRP
            data (list[int]): Data part

        Returns:
            list[int]: Computed checksum
        """
        encoding = (Bech32Encodings.BECH32
                    if data[0] == SegwitBech32Const.WITNESS_VER_BECH32
                    else Bech32Encodings.BECH32M)

        return Bech32Utils.ComputeChecksum(hrp, data, encoding)


class SegwitBech32Decoder(Bech32DecoderBase):
    """
    Segwit Bech32 decoder class.
    It provides methods for decoding Segwit Bech32 format.
    """

    @classmethod
    def Decode(cls,
               hrp: str,
               addr: str) -> Tuple[int, bytes]:
        """
        Decode from Segwit Bech32.

        Args:
            hrp (str) : Human readable part
            addr (str): Address

        Returns:
            tuple[int, bytes]: Witness version (index 0) and witness program (index 1)

        Raises:
            Bech32ChecksumError: If the checksum is not valid
            ValueError: If the bech32 string is not valid
        """

        # Decode string
        hrp_got, data = cls._DecodeBech32(addr,
                                          SegwitBech32Const.SEPARATOR,
                                          SegwitBech32Const.CHECKSUM_STR_LEN)
        # Check HRP
        if hrp != hrp_got:
            raise ValueError(
                f"Invalid format (HRP not valid, expected {hrp}, got {hrp_got})"
            )

        # Convert back from base32 (remove witness version)
        conv_data = Bech32BaseUtils.ConvertFromBase32(data[1:])

        # Check data length
        if (len(conv_data) < SegwitBech32Const.WITNESS_PROG_MIN_BYTE_LEN
                or len(conv_data) > SegwitBech32Const.WITNESS_PROG_MAX_BYTE_LEN):
            raise ValueError(f"Invalid format (witness program length not valid: {len(conv_data)})")
        # Check witness version
        wit_ver = data[0]
        if wit_ver > SegwitBech32Const.WITNESS_VER_MAX_VAL:
            raise ValueError(f"Invalid format (witness version not valid: {wit_ver})")
        if wit_ver == 0 and not len(conv_data) in SegwitBech32Const.WITNESS_VER_ZERO_DATA_BYTE_LEN:
            raise ValueError(f"Invalid format (length not valid: {len(conv_data)})")

        return wit_ver, BytesUtils.FromList(conv_data)

    @staticmethod
    def _VerifyChecksum(hrp: str,
                        data: List[int]) -> bool:
        """
        Verify the checksum from the specified HRP and converted data characters.

        Args:
            hrp  (str)      : HRP
            data (list[int]): Data part

        Returns:
            bool: True if valid, false otherwise
        """
        encoding = (Bech32Encodings.BECH32
                    if data[0] == SegwitBech32Const.WITNESS_VER_BECH32
                    else Bech32Encodings.BECH32M)

        return Bech32Utils.VerifyChecksum(hrp, data, encoding)



# Copyright (c) 2021 Emanuele Bellocchia
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""Module with BIP32 constants."""

# Imports
from .bip32_key_net_ver import Bip32KeyNetVersions


class Bip32Const:
    """Class container for BIP32 constants."""

    # Main net key net version (xpub / xprv)
    MAIN_NET_KEY_NET_VERSIONS: Bip32KeyNetVersions = Bip32KeyNetVersions(b"\x04\x88\xb2\x1e", b"\x04\x88\xad\xe4")
    # Test net key net version (tpub / tprv)
    TEST_NET_KEY_NET_VERSIONS: Bip32KeyNetVersions = Bip32KeyNetVersions(b"\x04\x35\x87\xcf", b"\x04\x35\x83\x94")
    # Key net version for BIP32 Kholaw that uses 64-byte private keys (xpub / xprv)
    # KHOLAW_KEY_NET_VERSIONS: Bip32KeyNetVersions = Bip32KeyNetVersions(b"\x04\x88\xb2\x1e", b"\x0f\x43\x31\xd4")



"""A convenience which constructs expression trees from an easy-to-read syntax

Use this unless you have a compelling reason not to; it performs some
optimizations that would be tedious to do when constructing an expression tree
by hand.

"""
from collections import OrderedDict

from .exceptions import BadGrammar, UndefinedLabel
from .expressions import (Literal, Regex, Sequence, OneOf,
    Lookahead, Optional, ZeroOrMore, OneOrMore, Not, TokenMatcher,
    expression, is_callable)
from .nodes import NodeVisitor
from .utils import evaluate_string

class Grammar(OrderedDict):
    """A collection of rules that describe a language

    You can start parsing from the default rule by calling ``parse()``
    directly on the ``Grammar`` object::

        g = Grammar('''
                    polite_greeting = greeting ", my good " title
                    greeting        = "Hi" / "Hello"
                    title           = "madam" / "sir"
                    ''')
        g.parse('Hello, my good sir')

    Or start parsing from any of the other rules; you can pull them out of the
    grammar as if it were a dictionary::

        g['title'].parse('sir')

    You could also just construct a bunch of ``Expression`` objects yourself
    and stitch them together into a language, but using a ``Grammar`` has some
    important advantages:

    * Languages are much easier to define in the nice syntax it provides.
    * Circular references aren't a pain.
    * It does all kinds of whizzy space- and time-saving optimizations, like
      factoring up repeated subexpressions into a single object, which should
      increase cache hit ratio. [Is this implemented yet?]

    """
    def __init__(self, rules='', **more_rules):
        """Construct a grammar.

        :arg rules: A string of production rules, one per line.
        :arg default_rule: The name of the rule invoked when you call
            :meth:`parse()` or :meth:`match()` on the grammar. Defaults to the
            first rule. Falls back to None if there are no string-based rules
            in this grammar.
        :arg more_rules: Additional kwargs whose names are rule names and
            values are Expressions or custom-coded callables which accomplish
            things the built-in rule syntax cannot. These take precedence over
            ``rules`` in case of naming conflicts.

        """

        decorated_custom_rules = {
            k: (expression(v, k, self) if is_callable(v) else v)
            for k, v in more_rules.items()}

        exprs, first = self._expressions_from_rules(rules, decorated_custom_rules)
        super(Grammar, self).__init__(exprs.items())
        self.default_rule = first  # may be None

    def default(self, rule_name):
        """Return a new Grammar whose :term:`default rule` is ``rule_name``."""
        new = self._copy()
        new.default_rule = new[rule_name]
        return new

    def _copy(self):
        """Return a shallow copy of myself.

        Deep is unnecessary, since Expression trees are immutable. Subgrammars
        recreate all the Expressions from scratch, and AbstractGrammars have
        no Expressions.

        """
        new = Grammar.__new__(Grammar)
        super(Grammar, new).__init__(self.items())
        new.default_rule = self.default_rule
        return new

    def _expressions_from_rules(self, rules, custom_rules):
        """Return a 2-tuple: a dict of rule names pointing to their
        expressions, and then the first rule.

        It's a web of expressions, all referencing each other. Typically,
        there's a single root to the web of references, and that root is the
        starting symbol for parsing, but there's nothing saying you can't have
        multiple roots.

        :arg custom_rules: A map of rule names to custom-coded rules:
            Expressions

        """
        tree = rule_grammar.parse(rules)
        return RuleVisitor(custom_rules).visit(tree)

    def parse(self, text, pos=0):
        """Parse some text with the :term:`default rule`.

        :arg pos: The index at which to start parsing

        """
        self._check_default_rule()
        return self.default_rule.parse(text, pos=pos)

    def match(self, text, pos=0):
        """Parse some text with the :term:`default rule` but not necessarily
        all the way to the end.

        :arg pos: The index at which to start parsing

        """
        self._check_default_rule()
        return self.default_rule.match(text, pos=pos)

    def _check_default_rule(self):
        """Raise RuntimeError if there is no default rule defined."""
        if not self.default_rule:
            raise RuntimeError("Can't call parse() on a Grammar that has no "
                               "default rule. Choose a specific rule instead, "
                               "like some_grammar['some_rule'].parse(...).")

    def __str__(self):
        """Return a rule string that, when passed to the constructor, would
        reconstitute the grammar."""
        exprs = [self.default_rule] if self.default_rule else []
        exprs.extend(expr for expr in self.values() if
                     expr is not self.default_rule)
        return '\n'.join(expr.as_rule() for expr in exprs)

    def __repr__(self):
        """Return an expression that will reconstitute the grammar."""
        return "Grammar({!r})".format(str(self))


class TokenGrammar(Grammar):
    """A Grammar which takes a list of pre-lexed tokens instead of text

    This is useful if you want to do the lexing yourself, as a separate pass:
    for example, to implement indentation-based languages.

    """
    def _expressions_from_rules(self, rules, custom_rules):
        tree = rule_grammar.parse(rules)
        return TokenRuleVisitor(custom_rules).visit(tree)


class BootstrappingGrammar(Grammar):
    """The grammar used to recognize the textual rules that describe other
    grammars

    This grammar gets its start from some hard-coded Expressions and claws its
    way from there to an expression tree that describes how to parse the
    grammar description syntax.

    """
    def _expressions_from_rules(self, rule_syntax, custom_rules):
        """Return the rules for parsing the grammar definition syntax.

        Return a 2-tuple: a dict of rule names pointing to their expressions,
        and then the top-level expression for the first rule.

        """
        # Hard-code enough of the rules to parse the grammar that describes the
        # grammar description language, to bootstrap:
        comment = Regex(r'#[^\r\n]*', name='comment')
        meaninglessness = OneOf(Regex(r'\s+'), comment, name='meaninglessness')
        _ = ZeroOrMore(meaninglessness, name='_')
        equals = Sequence(Literal('='), _, name='equals')
        label = Sequence(Regex(r'[a-zA-Z_][a-zA-Z_0-9]*'), _, name='label')
        reference = Sequence(label, Not(equals), name='reference')
        quantifier = Sequence(Regex(r'[*+?]'), _, name='quantifier')
        # This pattern supports empty literals. TODO: A problem?
        spaceless_literal = Regex(r'u?r?"[^"\\]*(?:\\.[^"\\]*)*"',
                                  ignore_case=True,
                                  dot_all=True,
                                  name='spaceless_literal')
        literal = Sequence(spaceless_literal, _, name='literal')
        regex = Sequence(Literal('~'),
                         literal,
                         Regex('[ilmsuxa]*', ignore_case=True),
                         _,
                         name='regex')
        atom = OneOf(reference, literal, regex, name='atom')
        quantified = Sequence(atom, quantifier, name='quantified')

        term = OneOf(quantified, atom, name='term')
        not_term = Sequence(Literal('!'), term, _, name='not_term')
        term.members = (not_term,) + term.members

        sequence = Sequence(term, OneOrMore(term), name='sequence')
        or_term = Sequence(Literal('/'), _, term, name='or_term')
        ored = Sequence(term, OneOrMore(or_term), name='ored')
        expression = OneOf(ored, sequence, term, name='expression')
        rule = Sequence(label, equals, expression, name='rule')
        rules = Sequence(_, OneOrMore(rule), name='rules')

        # Use those hard-coded rules to parse the (more extensive) rule syntax.
        # (For example, unless I start using parentheses in the rule language
        # definition itself, I should never have to hard-code expressions for
        # those above.)

        rule_tree = rules.parse(rule_syntax)

        # Turn the parse tree into a map of expressions:
        return RuleVisitor().visit(rule_tree)


# The grammar for parsing PEG grammar definitions:
# This is a nice, simple grammar. We may someday add to it, but it's a safe bet
# that the future will always be a superset of this.
rule_syntax = (r'''
    # Ignored things (represented by _) are typically hung off the end of the
    # leafmost kinds of nodes. Literals like "/" count as leaves.

    rules = _ rule*
    rule = label equals expression
    equals = "=" _
    literal = spaceless_literal _

    # So you can't spell a regex like `~"..." ilm`:
    spaceless_literal = ~"u?r?\"[^\"\\\\]*(?:\\\\.[^\"\\\\]*)*\""is /
                        ~"u?r?'[^'\\\\]*(?:\\\\.[^'\\\\]*)*'"is

    expression = ored / sequence / term
    or_term = "/" _ term
    ored = term or_term+
    sequence = term term+
    not_term = "!" term _
    lookahead_term = "&" term _
    term = not_term / lookahead_term / quantified / atom
    quantified = atom quantifier
    atom = reference / literal / regex / parenthesized
    regex = "~" spaceless_literal ~"[ilmsuxa]*"i _
    parenthesized = "(" _ expression ")" _
    quantifier = ~"[*+?]" _
    reference = label !equals

    # A subsequent equal sign is the only thing that distinguishes a label
    # (which begins a new rule) from a reference (which is just a pointer to a
    # rule defined somewhere else):
    label = ~"[a-zA-Z_][a-zA-Z_0-9]*" _

    # _ = ~r"\s*(?:#[^\r\n]*)?\s*"
    _ = meaninglessness*
    meaninglessness = ~r"\s+" / comment
    comment = ~r"#[^\r\n]*"
    ''')


class LazyReference(str):
    """A lazy reference to a rule, which we resolve after grokking all the
    rules"""

    name = u''

    # Just for debugging:
    def _as_rhs(self):
        return u'<LazyReference to %s>' % self


class RuleVisitor(NodeVisitor):
    """Turns a parse tree of a grammar definition into a map of ``Expression``
    objects

    This is the magic piece that breathes life into a parsed bunch of parse
    rules, allowing them to go forth and parse other things.

    """
    quantifier_classes = {'?': Optional, '*': ZeroOrMore, '+': OneOrMore}

    visit_expression = visit_term = visit_atom = NodeVisitor.lift_child

    def __init__(self, custom_rules=None):
        """Construct.

        :arg custom_rules: A dict of {rule name: expression} holding custom
            rules which will take precedence over the others

        """
        self.custom_rules = custom_rules or {}

    def visit_parenthesized(self, node, parenthesized):
        """Treat a parenthesized subexpression as just its contents.

        Its position in the tree suffices to maintain its grouping semantics.

        """
        left_paren, _, expression, right_paren, _ = parenthesized
        return expression

    def visit_quantifier(self, node, quantifier):
        """Turn a quantifier into just its symbol-matching node."""
        symbol, _ = quantifier
        return symbol

    def visit_quantified(self, node, quantified):
        atom, quantifier = quantified
        return self.quantifier_classes[quantifier.text](atom)

    def visit_lookahead_term(self, node, lookahead_term):
        ampersand, term, _ = lookahead_term
        return Lookahead(term)

    def visit_not_term(self, node, not_term):
        exclamation, term, _ = not_term
        return Not(term)

    def visit_rule(self, node, rule):
        """Assign a name to the Expression and return it."""
        label, equals, expression = rule
        expression.name = label  # Assign a name to the expr.
        return expression

    def visit_sequence(self, node, sequence):
        """A parsed Sequence looks like [term node, OneOrMore node of
        ``another_term``s]. Flatten it out."""
        term, other_terms = sequence
        return Sequence(term, *other_terms)

    def visit_ored(self, node, ored):
        first_term, other_terms = ored
        return OneOf(first_term, *other_terms)

    def visit_or_term(self, node, or_term):
        """Return just the term from an ``or_term``.

        We already know it's going to be ored, from the containing ``ored``.

        """
        slash, _, term = or_term
        return term

    def visit_label(self, node, label):
        """Turn a label into a unicode string."""
        name, _ = label
        return name.text

    def visit_reference(self, node, reference):
        """Stick a :class:`LazyReference` in the tree as a placeholder.

        We resolve them all later.

        """
        label, not_equals = reference
        return LazyReference(label)

    def visit_regex(self, node, regex):
        """Return a ``Regex`` expression."""
        tilde, literal, flags, _ = regex
        flags = flags.text.upper()
        pattern = literal.literal  # Pull the string back out of the Literal
                                   # object.
        return Regex(pattern, ignore_case='I' in flags,
                              locale='L' in flags,
                              multiline='M' in flags,
                              dot_all='S' in flags,
                              unicode='U' in flags,
                              verbose='X' in flags,
                              ascii='A' in flags)

    def visit_spaceless_literal(self, spaceless_literal, visited_children):
        """Turn a string literal into a ``Literal`` that recognizes it."""
        return Literal(evaluate_string(spaceless_literal.text))

    def visit_literal(self, node, literal):
        """Pick just the literal out of a literal-and-junk combo."""
        spaceless_literal, _ = literal
        return spaceless_literal

    def generic_visit(self, node, visited_children):
        """Replace childbearing nodes with a list of their children; keep
        others untouched.

        For our case, if a node has children, only the children are important.
        Otherwise, keep the node around for (for example) the flags of the
        regex rule. Most of these kept-around nodes are subsequently thrown
        away by the other visitor methods.

        We can't simply hang the visited children off the original node; that
        would be disastrous if the node occurred in more than one place in the
        tree.

        """
        return visited_children or node  # should semantically be a tuple

    def _resolve_refs(self, rule_map, expr, done):
        """Return an expression with all its lazy references recursively
        resolved.

        Resolve any lazy references in the expression ``expr``, recursing into
        all subexpressions.

        :arg done: The set of Expressions that have already been or are
            currently being resolved, to ward off redundant work and prevent
            infinite recursion for circular refs

        """
        if isinstance(expr, LazyReference):
            label = str(expr)
            try:
                reffed_expr = rule_map[label]
            except KeyError:
                raise UndefinedLabel(expr)
            return self._resolve_refs(rule_map, reffed_expr, done)
        else:
            if getattr(expr, 'members', ()) and expr not in done:
                # Prevents infinite recursion for circular refs. At worst, one
                # of `expr.members` can refer back to `expr`, but it can't go
                # any farther.
                done.add(expr)
                expr.members = tuple(self._resolve_refs(rule_map, member, done)
                                     for member in expr.members)
            return expr

    def visit_rules(self, node, rules_list):
        """Collate all the rules into a map. Return (map, default rule).

        The default rule is the first one. Or, if you have more than one rule
        of that name, it's the last-occurring rule of that name. (This lets you
        override the default rule when you extend a grammar.) If there are no
        string-based rules, the default rule is None, because the custom rules,
        due to being kwarg-based, are unordered.

        """
        _, rules = rules_list

        # Map each rule's name to its Expression. Later rules of the same name
        # override earlier ones. This lets us define rules multiple times and
        # have the last declaration win, so you can extend grammars by
        # concatenation.
        rule_map = OrderedDict((expr.name, expr) for expr in rules)

        # And custom rules override string-based rules. This is the least
        # surprising choice when you compare the dict constructor:
        # dict({'x': 5}, x=6).
        rule_map.update(self.custom_rules)

        # Resolve references. This tolerates forward references.
        done = set()
        rule_map = OrderedDict((expr.name, self._resolve_refs(rule_map, expr, done))
                               for expr in rule_map.values())

        # isinstance() is a temporary hack around the fact that * rules don't
        # always get transformed into lists by NodeVisitor. We should fix that;
        # it's surprising and requires writing lame branches like this.
        return rule_map, (rule_map[rules[0].name]
                          if isinstance(rules, list) and rules else None)


class TokenRuleVisitor(RuleVisitor):
    """A visitor which builds expression trees meant to work on sequences of
    pre-lexed tokens rather than strings"""

    def visit_spaceless_literal(self, spaceless_literal, visited_children):
        """Turn a string literal into a ``TokenMatcher`` that matches
        ``Token`` objects by their ``type`` attributes."""
        return TokenMatcher(evaluate_string(spaceless_literal.text))

    def visit_regex(self, node, regex):
        tilde, literal, flags, _ = regex
        raise BadGrammar('Regexes do not make sense in TokenGrammars, since '
                         'TokenGrammars operate on pre-lexed tokens rather '
                         'than characters.')


# Bootstrap to level 1...
rule_grammar = BootstrappingGrammar(rule_syntax)
# ...and then to level 2. This establishes that the node tree of our rule
# syntax is built by the same machinery that will build trees of our users'
# grammars. And the correctness of that tree is tested, indirectly, in
# test_grammar.
rule_grammar = Grammar(rule_syntax)


# TODO: Teach Expression trees how to spit out Python representations of
# themselves. Then we can just paste that in above, and we won't have to
# bootstrap on import. Though it'll be a little less DRY. [Ah, but this is not
# so clean, because it would have to output multiple statements to get multiple
# refs to a single expression hooked up.]
